@@ -12,22 +12,24 @@ import time
 import typing
 from typing import Any
 from typing import Mapping
+import warnings
 
 from ..api import CacheBackend
 from ..api import NO_VALUE
 from ... import util
 
+
 if typing.TYPE_CHECKING:
+    import bmemcached
     import memcache
     import pylibmc
-    import bmemcached
     import pymemcache
 else:
     # delayed import
-    memcache = None
-    pylibmc = None
-    bmemcached = None
-    pymemcache = None
+    bmemcached = None  # noqa F811
+    memcache = None  # noqa F811
+    pylibmc = None  # noqa F811
+    pymemcache = None  # noqa F811
 
 __all__ = (
     "GenericMemcachedBackend",
@@ -234,9 +236,6 @@ class MemcacheArgs(GenericMemcachedBackend):
         super(MemcacheArgs, self).__init__(arguments)
 
 
-pylibmc = None
-
-
 class PylibmcBackend(MemcacheArgs, GenericMemcachedBackend):
     """A backend for the
     `pylibmc <http://sendapatch.se/projects/pylibmc/index.html>`_
@@ -286,9 +285,6 @@ class PylibmcBackend(MemcacheArgs, GenericMemcachedBackend):
         )
 
 
-memcache = None
-
-
 class MemcachedBackend(MemcacheArgs, GenericMemcachedBackend):
     """A backend using the standard
     `Python-memcached <http://www.tummy.com/Community/software/\
@@ -317,9 +313,6 @@ class MemcachedBackend(MemcacheArgs, GenericMemcachedBackend):
         return memcache.Client(self.url)
 
 
-bmemcached = None
-
-
 class BMemcachedBackend(GenericMemcachedBackend):
     """A backend for the
     `python-binary-memcached <https://github.com/jaysonsantos/\
@@ -332,16 +325,6 @@ class BMemcachedBackend(GenericMemcachedBackend):
     SASL is a standard for adding authentication mechanisms
     to protocols in a way that is protocol independent.
 
-    SSL/TLS is a security layer on end-to-end communication.
-    It provides following benefits:
-
-    * Encryption: Data is encrypted on the wire between
-      Memcached client and server.
-    * Authentication: Optionally, both server and client
-      authenticate each other.
-    * Integrity: Data is not tampered or altered when
-      transmitted between client and server
-
     A typical configuration using username/password::
 
         from dogpile.cache import make_region
@@ -430,9 +413,6 @@ class BMemcachedBackend(GenericMemcachedBackend):
             self.delete(key)
 
 
-pymemcache = None
-
-
 class PyMemcacheBackend(GenericMemcachedBackend):
     """A backend for the
     `pymemcache <https://github.com/pinterest/pymemcache>`_
@@ -453,27 +433,17 @@ class PyMemcacheBackend(GenericMemcachedBackend):
       cache misses.
 
     dogpile.cache uses the ``HashClient`` from pymemcache in order to reduce
-    API differences when compared to other memcached client drivers. In short,
-    this allows the user to provide a single server or a list of memcached
+    API differences when compared to other memcached client drivers.
+    This allows the user to provide a single server or a list of memcached
     servers.
 
-    The ``serde`` param defaults to ``pymemcache.serde.pickle_serde`` as the
-    legacy ``serde`` would always convert the stored data to binary.
+    Arguments which can be passed to the ``arguments``
+    dictionary include:
 
-    The ``default_noreply`` param defaults to False, otherwise the add command
-    would always return True causing the mutex not to work.
+    :param tls_context: optional TLS context, will be used for
+     TLS connections.
 
-    SSL/TLS is a security layer on end-to-end communication.
-    It provides following benefits:
-
-    * Encryption: Data is encrypted on the wire between
-      Memcached client and server.
-    * Authentication: Optionally, both server and client
-      authenticate each other.
-    * Integrity: Data is not tampered or altered when
-      transmitted between client and server
-
-    A typical configuration using tls_context::
+     A typical configuration using tls_context::
 
         import ssl
         from dogpile.cache import make_region
@@ -489,19 +459,85 @@ class PyMemcacheBackend(GenericMemcachedBackend):
             }
         )
 
-    For advanced ways to configure TLS creating a more complex
-    tls_context visit https://docs.python.org/3/library/ssl.html
+     .. seealso::
 
-    Arguments which can be passed to the ``arguments``
-    dictionary include:
+        `<https://docs.python.org/3/library/ssl.html>`_ - additional TLS
+        documentation.
 
-    :param tls_context: optional TLS context, will be used for
-     TLS connections.
     :param serde: optional "serde". Defaults to
-     ``pymemcache.serde.pickle_serde``
-    :param default_noreply: Defaults to False
+     ``pymemcache.serde.pickle_serde``.
 
-    """
+    :param default_noreply:  defaults to False.  When set to True this flag
+     enables the pymemcache "noreply" feature.  See the pymemcache
+     documentation for further details.
+
+    :param socket_keepalive: optional socket keepalive, will be used for
+     TCP keepalive configuration.  Use of this parameter requires pymemcache
+     3.5.0 or greater.  This parameter
+     accepts a
+     `pymemcache.client.base.KeepAliveOpts
+     <https://pymemcache.readthedocs.io/en/latest/apidoc/pymemcache.client.base.html#pymemcache.client.base.KeepaliveOpts>`_
+     object.
+
+     A typical configuration using ``socket_keepalive``::
+
+        from pymemcache import KeepaliveOpts
+        from dogpile.cache import make_region
+
+        # Using the default keepalive configuration
+        socket_keepalive = KeepaliveOpts()
+
+        region = make_region().configure(
+            'dogpile.cache.pymemcache',
+            expiration_time = 3600,
+            arguments = {
+                'url':["127.0.0.1"],
+                'socket_keepalive': socket_keepalive
+            }
+        )
+
+     .. versionadded:: 1.1.4 - added support for ``socket_keepalive``.
+
+    :param enable_retry_client: optional flag to enable retry client
+     mechanisms to handle failure.  Defaults to False.  When set to ``True``,
+     the :paramref:`.PyMemcacheBackend.retry_attempts` parameter must also
+     be set, along with optional parameters
+     :paramref:`.PyMemcacheBackend.retry_delay`.
+     :paramref:`.PyMemcacheBackend.retry_for`,
+     :paramref:`.PyMemcacheBackend.do_not_retry_for`.
+
+     .. seealso::
+
+        `<https://pymemcache.readthedocs.io/en/latest/getting_started.html#using-the-built-in-retrying-mechanism>`_ -
+        in the pymemcache documentation
+
+     .. versionadded:: 1.1.4
+
+    :param retry_attempts: how many times to attempt an action before
+     failing. Must be 1 or above. Defaults to None.
+
+     .. versionadded:: 1.1.4
+
+    :param retry_delay: optional int|float, how many seconds to sleep between
+     each attempt. Defaults to None.
+
+     .. versionadded:: 1.1.4
+
+    :param retry_for: optional None|tuple|set|list, what exceptions to
+     allow retries for. Will allow retries for all exceptions if None.
+     Example: ``(MemcacheClientError, MemcacheUnexpectedCloseError)``
+     Accepts any class that is a subclass of Exception.  Defaults to None.
+
+     .. versionadded:: 1.1.4
+
+    :param do_not_retry_for: optional None|tuple|set|list, what
+     exceptions should be retried. Will not block retries for any Exception if
+     None. Example: ``(IOError, MemcacheIllegalInputError)``
+     Accepts any class that is a subclass of Exception. Defaults to None.
+
+     .. versionadded:: 1.1.4
+
+    """  # noqa E501
 
     def __init__(self, arguments):
         super().__init__(arguments)
@@ -509,15 +545,45 @@ class PyMemcacheBackend(GenericMemcachedBackend):
         self.serde = arguments.get("serde", pymemcache.serde.pickle_serde)
         self.default_noreply = arguments.get("default_noreply", False)
         self.tls_context = arguments.get("tls_context", None)
+        self.socket_keepalive = arguments.get("socket_keepalive", None)
+        self.enable_retry_client = arguments.get("enable_retry_client", False)
+        self.retry_attempts = arguments.get("retry_attempts", None)
+        self.retry_delay = arguments.get("retry_delay", None)
+        self.retry_for = arguments.get("retry_for", None)
+        self.do_not_retry_for = arguments.get("do_not_retry_for", None)
+        if (
+            self.retry_delay is not None
+            or self.retry_attempts is not None
+            or self.retry_for is not None
+            or self.do_not_retry_for is not None
+        ) and not self.enable_retry_client:
+            warnings.warn(
+                "enable_retry_client is not set; retry options "
+                "will be ignored"
+            )
 
     def _imports(self):
         global pymemcache
         import pymemcache
 
     def _create_client(self):
-        return pymemcache.client.hash.HashClient(
-            self.url,
-            serde=self.serde,
-            default_noreply=self.default_noreply,
-            tls_context=self.tls_context,
-        )
+        _kwargs = {
+            "serde": self.serde,
+            "default_noreply": self.default_noreply,
+            "tls_context": self.tls_context,
+        }
+        if self.socket_keepalive is not None:
+            _kwargs.update({"socket_keepalive": self.socket_keepalive})
+
+        client = pymemcache.client.hash.HashClient(self.url, **_kwargs)
+
+        if self.enable_retry_client:
+            return pymemcache.client.retrying.RetryingClient(
+                client,
+                attempts=self.retry_attempts,
+                retry_delay=self.retry_delay,
+                retry_for=self.retry_for,
+                do_not_retry_for=self.do_not_retry_for,
+            )
+
+        return client
